@@ -81,17 +81,19 @@ class MessageOutput:
                     self.comments[t] = comment
 
     def outputHeader(self, out):
+        import time
         out.write("""msgid ""
 msgstr ""
 "Project-Id-Version: PACKAGE VERSION\\n"
-"PO-Revision-Date: YEAR-MO-DA HO:MI +ZONE\\n"
-"Language-Team: LANGUAGE\\n"
-"Last-Translator: TRANSLATOR\\n"
+"POT-Creation-Date: %s\\n"
+"PO-Revision-Date: YEAR-MO-DA HO:MI+ZONE\\n"
+"Last-Translator: FULL NAME <EMAIL@ADDRESS>\\n"
+"Language-Team: LANGUAGE <LL@li.org>\\n"
 "MIME-Version: 1.0\\n"
 "Content-Type: text/plain; charset=UTF-8\\n"
 "Content-Transfer-Encoding: 8bit\\n"
 
-""")
+""" % (time.strftime("%Y-%m-%d %H:%M%z")))
 
     def outputAll(self, out):
         self.outputHeader(out)
@@ -114,7 +116,6 @@ msgstr ""
 
 
 def normalizeNode(node):
-    #print node.content
     if not node:
         return
     elif isSpacePreserveNode(node):
@@ -125,7 +126,7 @@ def normalizeNode(node):
         else:
             node.setContent(re.sub('\s+',' ', node.content))
 
-    elif node.children:
+    elif node.children and node.type == 'element':
         child = node.children
         while child:
             normalizeNode(child)
@@ -140,16 +141,17 @@ def normalizeString(text, ignorewhitespace = 1):
     try:
         # Lets add document DTD so entities are resolved
         dtd = doc.intSubset()
-        tmp = ''
-        if expand_entities: # FIXME: we get a "Segmentation fault" in libxml2.parseMemory() when we include DTD otherwise
-            tmp = dtd.serialize()
+        tmp = dtd.serialize('utf-8')
         tmp = tmp + '<norm>%s</norm>' % text
     except:
         tmp = '<norm>%s</norm>' % text
 
     try:
-        #libxml2.replaceEntities(0)
-        tree = libxml2.parseMemory(tmp,len(tmp))
+        ctxt = libxml2.createDocParserCtxt(tmp)
+        if expand_entities:
+            ctxt.replaceEntities(1)
+        ctxt.parseDocument()
+        tree = ctxt.doc()
         newnode = tree.getRootElement()
     except:
         print >> sys.stderr, """Error while normalizing string as XML:\n"%s"\n""" % (text)
@@ -170,11 +172,11 @@ def normalizeString(text, ignorewhitespace = 1):
 
 def stringForEntity(node):
     """Replaces entities in the node."""
-    text = node.serialize()
+    text = node.serialize('utf-8')
     try:
         # Lets add document DTD so entities are resolved
         dtd = node.doc.intSubset()
-        tmp = dtd.serialize() + '<norm>%s</norm>' % text
+        tmp = dtd.serialize('utf-8') + '<norm>%s</norm>' % text
         next = 1
     except:
         tmp = '<norm>%s</norm>' % text
@@ -218,7 +220,7 @@ def getTranslation(text, spacepreserve = 0):
     if file:
         gt = gettext.GNUTranslations(file)
         if gt:
-            return gt.gettext(text)
+            return gt.ugettext(text.decode('utf-8'))
     return text
 
 def startTagForNode(node):
@@ -231,12 +233,16 @@ def startTagForNode(node):
         for p in node.properties:
             if p.type == 'attribute':
                 # FIXME: This part sucks
-                try:
-                    params += ' %s:%s="%s"' % (p.ns().name, p.name, p.content)
-                except:
-                    params += ' %s="%s"' % (p.name, p.content)
+                params += p.serialize('utf-8')
     return result+params
         
+def endTagForNode(node):
+    if not node:
+        return 0
+
+    result = node.name
+    return result
+        
 def isFinalNode(node):
     if automatic:
         auto = autoNodeIsFinal(node)
@@ -299,27 +305,26 @@ def getCommentForNode(node):
 
 def replaceNodeContentsWithText(node,text):
     """Replaces all subnodes of a node with contents of text treated as XML."""
-    #print >> sys.stderr, text
     if node.children:
-        starttag = startTagForNode(node)
-        endtag = node.name
+        starttag = node.name #startTagForNode(node)
+        endtag = endTagForNode(node)
         try:
             # Lets add document DTD so entities are resolved
             dtd = doc.intSubset()
             tmp = ''
             if expand_entities: # FIXME: we get a "Segmentation fault" in libxml2.parseMemory() when we include DTD otherwise
-                tmp = dtd.serialize()
+                tmp = dtd.serialize('utf-8')
             tmp = tmp + '<%s>%s</%s>' % (starttag, text, endtag)
         except:
             tmp = '<%s>%s</%s>' % (starttag, text, endtag)
 
         try:
-            ctxt = libxml2.createDocParserCtxt(tmp)
+            ctxt = libxml2.createDocParserCtxt(tmp.encode('utf-8'))
             ctxt.replaceEntities(0)
             ctxt.parseDocument()
             newnode = ctxt.doc()
         except:
-            print >> sys.stderr, """Error while parsing translation as XML:\n"%s"\n""" % (text)
+            print >> sys.stderr, """Error while parsing translation as XML:\n"%s"\n""" % (text.encode('utf-8'))
             return
 
         newelem = newnode.getRootElement()
@@ -360,9 +365,11 @@ def worthOutputting(node):
     """
     worth = 1
     parent = node.parent
-    final = isFinalNode(node)
+    final = isFinalNode(node) and node.name not in ignored_tags
     while not final and parent:
         if isFinalNode(parent):
+            final = 1 # reset if we've got to one final tag
+        if final and (parent.name not in ignored_tags) and worthOutputting(parent):
             worth = 0
             break
         parent = parent.parent
@@ -371,45 +378,52 @@ def worthOutputting(node):
 
     return autoNodeIsFinal(node)
     
-def processElementTag(node):
+def processElementTag(node, replacements, restart = 0):
     """Process node with node.type == 'element'."""
     if node.type == 'element':
         outtxt = ''
-
-        if not worthOutputting(node):
-            child = node.children
-            while child:
-                outtxt += doSerialize(child)
-                child = child.next
+        if restart:
+            myrepl = []
         else:
-            PlaceHolder = 0
-            submsgs = {}
+            myrepl = replacements
 
-            child = node.children
-            while child:
-                if isFinalNode(child):
-                    PlaceHolder += 1
-                    (starttag, submsg, endtag) = processElementTag(child)
-                    outtxt += '<placeholder-%d/>' % (PlaceHolder)
-                    if mode=='merge':
-                        submsgs[PlaceHolder] = (starttag, getTranslation(submsg, isSpacePreserveNode(node)), endtag)
+        submsgs = []
+
+        child = node.children
+        while child:
+            if (isFinalNode(child)) or (child.type == 'element' and worthOutputting(child)):
+                myrepl.append(processElementTag(child, myrepl, 1))
+                outtxt += '<placeholder-%d/>' % (len(myrepl))
+            else:
+                if child.type == 'element':
+                    (starttag, content, endtag, translation) = processElementTag(child, myrepl, 0)
+                    outtxt += '<%s>%s</%s>' % (starttag, content, endtag)
                 else:
                     outtxt += doSerialize(child)
 
-                child = child.next
+            child = child.next
 
-            if mode=='merge':
-                outtxt = getTranslation(outtxt, isSpacePreserveNode(node))
-                for i in submsgs.keys():
-                    outtxt = outtxt.replace('<placeholder-%d/>' % (i), ''.join(submsgs[i]))
-                if worthOutputting(node):
-                    replaceNodeContentsWithText(node, outtxt)
-            elif worthOutputting(node):
-                msg.outputMessage(outtxt, node.lineNo(), getCommentForNode(node), isSpacePreserveNode(node), tag = node.name)
+        if mode == 'merge':
+            translation = getTranslation(outtxt, isSpacePreserveNode(node))
+        else:
+            translation = outtxt
+        starttag = startTagForNode(node)
+        endtag = endTagForNode(node)
 
-        starttag = '<' + startTagForNode(node) + '>'
-        endtag = '</' + node.name + '>'
-        return (starttag, outtxt, endtag)
+        if restart or worthOutputting(node):
+            i = 0
+            while i < len(myrepl):
+                replacement = '<%s>%s</%s>' % (myrepl[i][0], myrepl[i][3], myrepl[i][2])
+                i += 1
+                translation = translation.replace('<placeholder-%d/>' % (i), replacement)
+
+            if worthOutputting(node):
+                if mode == 'merge':
+                    replaceNodeContentsWithText(node, translation)
+                else:
+                    msg.outputMessage(outtxt, node.lineNo(), getCommentForNode(node), isSpacePreserveNode(node), tag = node.name)
+
+        return (starttag, outtxt, endtag, translation)
     else:
         raise Exception("You must pass node with node.type=='element'.")
 
@@ -458,7 +472,9 @@ def doSerialize(node):
     elif node.type == 'text':
         return node.serialize('utf-8')
     elif node.type == 'element':
-        return ''.join(processElementTag(node))
+        repl = []
+        (starttag, content, endtag, translation) = processElementTag(node, repl, 1)
+        return '<%s>%s</%s>' % (starttag, content, endtag)
     else:
         child = node.children
         outtxt = ''
@@ -531,7 +547,7 @@ def tryToUpdate(allargs, lang):
             sys.stderr.write("Error: cannot rename file.\n")
             sys.exit(11)
         else:
-            os.system("msgfmt -cv -o /dev/null %s" % (file))
+            os.system("msgfmt -cv -o %s %s" % (NULL_STRING, file))
             sys.exit(0)
 
 def load_mode(modename):
@@ -565,6 +581,7 @@ mofile = ''
 ultimate = [ ]
 ignored = [ ]
 filenames = [ ]
+translationlanguage = ''
 
 mode = 'pot' # 'pot' or 'merge'
 automatic = 0
@@ -573,6 +590,9 @@ expand_all_entities = 0
 
 output  = '-' # this means to stdout
 
+NULL_STRING = '/dev/null'
+if not os.path.exists('/dev/null'): NULL_STRING = 'NUL'
+
 import getopt, fileinput
 
 def usage (with_help = False):
@@ -591,6 +611,7 @@ OPTIONS may be some of:
     -r    --reuse=FILE         Specify translated XML file with the same structure
     -t    --translation=FILE   Specify MO file containing translation, and merge
     -u    --update-translation=LANG.po   Updates a PO file using msgmerge program
+    -l    --language=LANG      Set language of the translation to LANG
     -v    --version            Output version of the xml2po program
 
     -h    --help               Output this message
@@ -610,9 +631,9 @@ EXAMPLES:
 if len(sys.argv) < 2: usage()
 
 args = sys.argv[1:]
-try: opts, args = getopt.getopt(args, 'avhkem:t:o:p:u:r:',
+try: opts, args = getopt.getopt(args, 'avhkem:t:o:p:u:r:l:',
                            ['automatic-tags','version', 'help', 'keep-entities', 'expand-all-entities', 'mode=', 'translation=',
-                            'output=', 'po-file=', 'update-translation=', 'reuse=' ])
+                            'output=', 'po-file=', 'update-translation=', 'reuse=', 'language=' ])
 except getopt.GetoptError: usage(True)
 
 for opt, arg in opts:
@@ -624,10 +645,12 @@ for opt, arg in opts:
         expand_entities = 0
     elif opt in ('-e', '--expand-all-entities'):
         expand_all_entities = 1
+    elif opt in ('-l', '--language'):
+        translationlanguage = arg
     elif opt in ('-t', '--translation'):
         mofile = arg
         mode = 'merge'
-        translationlanguage = os.path.splitext(mofile)[0]
+        if translationlanguage == '': translationlanguage = os.path.split(os.path.splitext(mofile)[0])[1]
     elif opt in ('-r', '--reuse'):
         origxml = arg
     elif opt in ('-u', '--update-translation'):
@@ -635,8 +658,8 @@ for opt, arg in opts:
     elif opt in ('-p', '--po-file'):
         mofile = ".xml2po.mo"
         pofile = arg
-        translationlanguage = os.path.splitext(pofile)[0]
-        os.system("msgfmt -o %s %s >/dev/null" % (mofile, pofile)) and sys.exit(7)
+        if translationlanguage == '': translationlanguage = os.path.split(os.path.splitext(pofile)[0])[1]
+        os.system("msgfmt -o %s %s >%s" % (mofile, pofile, NULL_STRING)) and sys.exit(7)
         mode = 'merge'
     elif opt in ('-o', '--output'):
         output = arg
